@@ -42,6 +42,7 @@
 #include <QPropertyAnimation>
 #include <QJsonArray>
 #include <QJsonObject>
+#include <QTextDocumentFragment>
 
 #include <KLocalizedString>
 #include <KColorScheme>
@@ -52,18 +53,6 @@ const QString CommandEntry::HidePrompt = QLatin1String(">   ");
 const double CommandEntry::HorizontalSpacing = 4;
 const double CommandEntry::VerticalSpacing = 4;
 
-static const int colorsCount = 26;
-static QColor colors[colorsCount] = {QColor(255,255,255), QColor(0,0,0),
-							QColor(192,0,0), QColor(255,0,0), QColor(255,192,192), //red
-							QColor(0,192,0), QColor(0,255,0), QColor(192,255,192), //green
-							QColor(0,0,192), QColor(0,0,255), QColor(192,192,255), //blue
-							QColor(192,192,0), QColor(255,255,0), QColor(255,255,192), //yellow
-							QColor(0,192,192), QColor(0,255,255), QColor(192,255,255), //cyan
-							QColor(192,0,192), QColor(255,0,255), QColor(255,192,255), //magenta
-							QColor(192,88,0), QColor(255,128,0), QColor(255,168,88), //orange
-							QColor(128,128,128), QColor(160,160,160), QColor(195,195,195) //grey
-							};
-
 
 CommandEntry::CommandEntry(Worksheet* worksheet) : WorksheetEntry(worksheet),
     m_promptItem(new WorksheetTextItem(this, Qt::NoTextInteraction)),
@@ -81,7 +70,8 @@ CommandEntry::CommandEntry(Worksheet* worksheet) : WorksheetEntry(worksheet),
     m_backgroundColorMenu(nullptr),
     m_textColorActionGroup(nullptr),
     m_textColorMenu(nullptr),
-    m_fontMenu(nullptr)
+    m_fontMenu(nullptr),
+    m_isExecutionEnabled(true)
 {
     m_promptItem->setPlainText(Prompt);
     m_promptItem->setItemDragable(true);
@@ -102,8 +92,8 @@ CommandEntry::CommandEntry(Worksheet* worksheet) : WorksheetEntry(worksheet),
 
     connect(&m_controlElement, &WorksheetControlItem::doubleClick, this, &CommandEntry::changeResultCollapsingAction);
 
-    connect(m_commandItem, &WorksheetTextItem::tabPressed, this, &CommandEntry::showCompletion);
-    connect(m_commandItem, &WorksheetTextItem::backtabPressed, this, &CommandEntry::selectPreviousCompletion);
+    connect(m_commandItem, &WorksheetTextItem::tabPressed, this, &CommandEntry::handleTabPress);
+    connect(m_commandItem, &WorksheetTextItem::backtabPressed, this, &CommandEntry::handleBacktabPress);
     connect(m_commandItem, &WorksheetTextItem::applyCompletion, this, &CommandEntry::applySelectedCompletion);
     connect(m_commandItem, &WorksheetTextItem::execute, this, [=]() { evaluate();} );
     connect(m_commandItem, &WorksheetTextItem::moveToPrevious, this, &CommandEntry::moveToPreviousItem);
@@ -134,22 +124,10 @@ int CommandEntry::type() const
 }
 
 void CommandEntry::initMenus() {
-    //background color
-	const QString colorNames[colorsCount] = {i18n("White"), i18n("Black"),
-							i18n("Dark Red"), i18n("Red"), i18n("Light Red"),
-							i18n("Dark Green"), i18n("Green"), i18n("Light Green"),
-							i18n("Dark Blue"), i18n("Blue"), i18n("Light Blue"),
-							i18n("Dark Yellow"), i18n("Yellow"), i18n("Light Yellow"),
-							i18n("Dark Cyan"), i18n("Cyan"), i18n("Light Cyan"),
-							i18n("Dark Magenta"), i18n("Magenta"), i18n("Light Magenta"),
-							i18n("Dark Orange"), i18n("Orange"), i18n("Light Orange"),
-							i18n("Dark Grey"), i18n("Grey"), i18n("Light Grey")
-							};
-
     //background color
     m_backgroundColorActionGroup = new QActionGroup(this);
-	m_backgroundColorActionGroup->setExclusive(true);
-	connect(m_backgroundColorActionGroup, &QActionGroup::triggered, this, &CommandEntry::backgroundColorChanged);
+    m_backgroundColorActionGroup->setExclusive(true);
+    connect(m_backgroundColorActionGroup, &QActionGroup::triggered, this, &CommandEntry::backgroundColorChanged);
 
     m_backgroundColorMenu = new QMenu(i18n("Background Color"));
     m_backgroundColorMenu->setIcon(QIcon::fromTheme(QLatin1String("format-fill-color")));
@@ -172,14 +150,15 @@ void CommandEntry::initMenus() {
         action->setCheckable(true);
         m_backgroundColorMenu->addAction(action);
 
-        if (m_backgroundColorCustom && m_commandItem->backgroundColor() == colors[i])
+        const QColor& backgroundColor = (m_isExecutionEnabled ? m_commandItem->backgroundColor() : m_activeExecutionBackgroundColor);
+        if (m_backgroundColorCustom && backgroundColor == colors[i])
             action->setChecked(true);
     }
 
-	//text color
+    //text color
     m_textColorActionGroup = new QActionGroup(this);
-	m_textColorActionGroup->setExclusive(true);
-	connect(m_textColorActionGroup, &QActionGroup::triggered, this, &CommandEntry::textColorChanged);
+    m_textColorActionGroup->setExclusive(true);
+    connect(m_textColorActionGroup, &QActionGroup::triggered, this, &CommandEntry::textColorChanged);
 
     m_textColorMenu = new QMenu(i18n("Text Color"));
     m_textColorMenu->setIcon(QIcon::fromTheme(QLatin1String("format-text-color")));
@@ -199,7 +178,8 @@ void CommandEntry::initMenus() {
         action->setCheckable(true);
         m_textColorMenu->addAction(action);
 
-        if (m_textColorCustom && m_commandItem->defaultTextColor() == colors[i])
+        const QColor& textColor = (m_isExecutionEnabled ? m_commandItem->defaultTextColor() : m_activeExecutionTextColor);
+        if (m_textColorCustom && textColor == colors[i])
             action->setChecked(true);
     }
 
@@ -248,13 +228,19 @@ void CommandEntry::backgroundColorChanged(QAction* action) {
     if (index == -1 || index>=colorsCount)
         index = 0;
 
+    QColor color;
     if (index == 0)
     {
         KColorScheme scheme = KColorScheme(QPalette::Normal, KColorScheme::View);
-        m_commandItem->setBackgroundColor(scheme.background(KColorScheme::AlternateBackground).color());
+        color = scheme.background(KColorScheme::AlternateBackground).color();
     }
     else
-        m_commandItem->setBackgroundColor(colors[index-1]);
+        color = colors[index-1];
+
+    if (m_isExecutionEnabled)
+        m_commandItem->setBackgroundColor(color);
+    else
+        m_activeExecutionBackgroundColor = color;
 }
 
 void CommandEntry::textColorChanged(QAction* action) {
@@ -262,12 +248,18 @@ void CommandEntry::textColorChanged(QAction* action) {
     if (index == -1 || index>=colorsCount)
         index = 0;
 
+    QColor color;
     if (index == 0)
     {
-        m_commandItem->setDefaultTextColor(m_defaultDefaultTextColor);
+        color = m_defaultDefaultTextColor;
     }
     else
-        m_commandItem->setDefaultTextColor(colors[index-1]);
+        color = colors[index-1];
+
+    if (m_isExecutionEnabled)
+        m_commandItem->setDefaultTextColor(color);
+    else
+        m_activeExecutionTextColor = color;
 }
 
 void CommandEntry::fontBoldTriggered()
@@ -358,6 +350,11 @@ void CommandEntry::populateMenu(QMenu* menu, QPointF pos)
             menu->addAction(i18n("Hide Results"), this, &CommandEntry::collapseResults);
     }
 
+    if (m_isExecutionEnabled)
+        menu->addAction(i18n("Exclude from Execution"), this, &CommandEntry::excludeFromExecution);
+    else
+        menu->addAction(i18n("Add to Execution"), this, &CommandEntry::addToExecution);
+
     menu->addMenu(m_backgroundColorMenu);
     menu->addMenu(m_textColorMenu);
     menu->addMenu(m_fontMenu);
@@ -519,6 +516,10 @@ void CommandEntry::setContent(const QDomElement& content, const KZip& file)
         }
     }
 
+    m_isExecutionEnabled = !(bool)(content.attribute(QLatin1String("ExecutionDisabled"), QLatin1String("0")).toInt());
+    if (m_isExecutionEnabled == false)
+        excludeFromExecution();
+
     setExpression(expr);
 }
 
@@ -610,42 +611,133 @@ QJsonValue CommandEntry::toJupyterJson()
     return entry;
 }
 
-void CommandEntry::showCompletion()
+void CommandEntry::handleTabPress()
 {
-    const QString line = currentLine();
+    const QString& line = currentLine();
 
-    if(!worksheet()->completionEnabled() || line.trimmed().isEmpty())
+    // When the auto completion is disabled, we insert 'Tab' and exit immediately
+    // When the auto completion is enabled, the logic is more complicated
+    if(!worksheet()->completionEnabled())
     {
         if (m_commandItem->hasFocus())
             m_commandItem->insertTab();
         return;
-    } else if (isShowingCompletionPopup()) {
-        QString comp = m_completionObject->completion();
-        qDebug() << "command" << m_completionObject->command();
-        qDebug() << "completion" << comp;
-        if (comp != m_completionObject->command()
-            || !m_completionObject->hasMultipleMatches()) {
-            if (m_completionObject->hasMultipleMatches()) {
-                completeCommandTo(comp, PreliminaryCompletion);
-            } else {
-                completeCommandTo(comp, FinalCompletion);
-                m_completionBox->hide();
-            }
-        } else {
-            m_completionBox->down();
-        }
-    } else {
+    }
+
+    if (isShowingCompletionPopup())
+        handleExistedCompletionBox();
+    else
+    {
+        QTextCursor cursor = m_commandItem->textCursor();
         int p = m_commandItem->textCursor().positionInBlock();
-        Cantor::CompletionObject* tco=worksheet()->session()->completionFor(line, p);
-        if(tco)
-            setCompletion(tco);
+
+        if (cursor.hasSelection())
+        {
+            int count = 1 + cursor.selectedText().count(QChar(0x2029));
+            cursor.setPosition(cursor.selectionEnd());
+            cursor.beginEditBlock();
+            for (int i = 0; i < count; i++)
+            {
+                cursor.movePosition(QTextCursor::StartOfLine);
+                cursor.insertText(QLatin1String("    "));
+                cursor.movePosition(QTextCursor::StartOfLine);
+                cursor.movePosition(QTextCursor::PreviousCharacter);
+            }
+            cursor.endEditBlock();
+        }
+        else if(line.left(p).trimmed().isEmpty())
+        {
+            if (m_commandItem->hasFocus())
+                m_commandItem->insertTab();
+        }
+        else
+            makeCompletion(line, p);
     }
 }
 
-void CommandEntry::selectPreviousCompletion()
+void CommandEntry::showCompletion()
 {
+    if(!worksheet()->completionEnabled())
+        return;
+
+    if (isShowingCompletionPopup())
+        handleExistedCompletionBox();
+    else
+    {
+        int p = m_commandItem->textCursor().positionInBlock();
+        makeCompletion(currentLine(), p);
+    }
+}
+
+void CommandEntry::handleExistedCompletionBox()
+{
+    QString comp = m_completionObject->completion();
+    if (comp != m_completionObject->command() || !m_completionObject->hasMultipleMatches())
+    {
+        if (m_completionObject->hasMultipleMatches())
+            completeCommandTo(comp, PreliminaryCompletion);
+        else
+        {
+            completeCommandTo(comp, FinalCompletion);
+            m_completionBox->hide();
+        }
+    }
+    else
+    {
+        m_completionBox->down();
+    }
+}
+
+void CommandEntry::makeCompletion(const QString& line, int position)
+{
+    Cantor::CompletionObject* tco=worksheet()->session()->completionFor(line, position);
+    if(tco)
+        setCompletion(tco);
+}
+
+void CommandEntry::handleBacktabPress()
+{
+    QTextCursor cursor = m_commandItem->textCursor();
+
     if (isShowingCompletionPopup())
         m_completionBox->up();
+    else if (cursor.hasSelection())
+    {
+        int count = 1 + cursor.selectedText().count(QChar(0x2029));
+        cursor.setPosition(cursor.selectionEnd());
+        cursor.beginEditBlock();
+        for (int i = 0; i < count; i++)
+        {
+            const QString& line = cursor.block().text();
+            cursor.movePosition(QTextCursor::StartOfLine);
+            int counter = 0;
+            // 4 spaces is Cantor tabulation size, so we remove also 4 characters or less
+            while (cursor.positionInBlock() < line.length() && line[cursor.positionInBlock()] == QLatin1Char(' ') && counter < 4)
+            {
+                cursor.movePosition(QTextCursor::NextCharacter, QTextCursor::KeepAnchor);
+                counter++;
+            }
+            cursor.removeSelectedText();
+            cursor.movePosition(QTextCursor::PreviousCharacter);
+        }
+        cursor.endEditBlock();
+    }
+    else
+    {
+        const QString& line = currentLine();
+        if (line.length() >= 4)
+        {
+            cursor.movePosition(QTextCursor::StartOfLine);
+            int counter = 0;
+            // 4 spaces is Cantor tabulation size, so we remove also 4 characters or less
+            while (cursor.positionInBlock() < line.length() && line[cursor.positionInBlock()] == QLatin1Char(' ') && counter < 4)
+            {
+                cursor.movePosition(QTextCursor::NextCharacter, QTextCursor::KeepAnchor);
+                counter++;
+            }
+            cursor.removeSelectedText();
+        }
+    }
 }
 
 QString CommandEntry::toPlain(const QString& commandSep, const QString& commentStartingSeq, const QString& commentEndingSeq)
@@ -665,6 +757,9 @@ QDomElement CommandEntry::toXml(QDomDocument& doc, KZip* archive)
     cmdElem.appendChild(doc.createTextNode( command() ));
     exprElem.appendChild(cmdElem);
 
+    if (!m_isExecutionEnabled)
+        exprElem.setAttribute(QLatin1String("ExecutionDisabled"), true);
+
     // save results of the expression if they exist
     if (expression())
     {
@@ -689,12 +784,12 @@ QDomElement CommandEntry::toXml(QDomDocument& doc, KZip* archive)
     // If user can change value from menu (menus have been inited) - check via menu
     // If use don't have menu, check if loaded color was custom color
     if (m_backgroundColorActionGroup)
-        isBackgroundColorNotDefault = m_backgroundColorActionGroup->checkedAction()->text() != i18n("Default");
+        isBackgroundColorNotDefault = m_backgroundColorActionGroup->actions().indexOf(m_backgroundColorActionGroup->checkedAction()) != 0;
     else
         isBackgroundColorNotDefault = m_backgroundColorCustom;
     if (isBackgroundColorNotDefault)
     {
-        QColor backgroundColor = m_commandItem->backgroundColor();
+        QColor backgroundColor = (m_isExecutionEnabled ? m_commandItem->backgroundColor() : m_activeExecutionBackgroundColor);
         QDomElement colorElem = doc.createElement( QLatin1String("Background") );
         colorElem.setAttribute(QLatin1String("red"), QString::number(backgroundColor.red()));
         colorElem.setAttribute(QLatin1String("green"), QString::number(backgroundColor.green()));
@@ -704,12 +799,12 @@ QDomElement CommandEntry::toXml(QDomDocument& doc, KZip* archive)
 
     //save the text properties if they differ from default values
     const QFont& font = m_commandItem->font();
-    const QColor& textColor = m_commandItem->defaultTextColor();
+    const QColor& textColor = (m_isExecutionEnabled ? m_commandItem->defaultTextColor() : m_activeExecutionTextColor);
     bool isFontNotDefault = font != QFontDatabase::systemFont(QFontDatabase::FixedFont);
 
     bool isTextColorNotDefault = false;
     if (m_textColorActionGroup)
-        isTextColorNotDefault = m_textColorActionGroup->checkedAction()->text() != i18n("Default");
+        isTextColorNotDefault = m_textColorActionGroup->actions().indexOf(m_textColorActionGroup->checkedAction()) != 0;
     else
         isTextColorNotDefault = m_textColorCustom;
 
@@ -770,33 +865,41 @@ bool CommandEntry::evaluateCurrentItem()
 
 bool CommandEntry::evaluate(EvaluationOption evalOp)
 {
-    if (worksheet()->session()->status() == Cantor::Session::Disable)
-        worksheet()->loginToSession();
+    if (m_isExecutionEnabled)
+    {
+        if (worksheet()->session()->status() == Cantor::Session::Disable)
+            worksheet()->loginToSession();
 
-    removeContextHelp();
-    QToolTip::hideText();
+        removeContextHelp();
+        QToolTip::hideText();
 
-    QString cmd = command();
-    m_evaluationOption = evalOp;
+        QString cmd = command();
+        m_evaluationOption = evalOp;
 
-    if(cmd.isEmpty()) {
-        removeResults();
-        for (auto* item : m_informationItems) {
-            item->deleteLater();
+        if(cmd.isEmpty()) {
+            removeResults();
+            for (auto* item : m_informationItems) {
+                item->deleteLater();
+            }
+            m_informationItems.clear();
+            recalculateSize();
+
+            evaluateNext(m_evaluationOption);
+            return false;
         }
-        m_informationItems.clear();
-        recalculateSize();
 
+        Cantor::Expression* expr = worksheet()->session()->evaluateExpression(cmd);
+        connect(expr, &Cantor::Expression::gotResult, this, [=]() { worksheet()->gotResult(expr); });
+
+        setExpression(expr);
+
+        return true;
+    }
+    else
+    {
         evaluateNext(m_evaluationOption);
-        return false;
+        return true;
     }
-
-    Cantor::Expression* expr = worksheet()->session()->evaluateExpression(cmd);
-    connect(expr, &Cantor::Expression::gotResult, this, [=]() { worksheet()->gotResult(expr); });
-
-    setExpression(expr);
-
-    return true;
 }
 
 void CommandEntry::interruptEvaluation()
@@ -1320,13 +1423,13 @@ WorksheetCursor CommandEntry::search(const QString& pattern, unsigned flags,
     return WorksheetCursor();
 }
 
-void CommandEntry::layOutForWidth(qreal w, bool force)
+void CommandEntry::layOutForWidth(qreal entry_zone_x, qreal w, bool force)
 {
-    if (w == size().width() && !force)
+    if (w == size().width() && m_commandItem->pos().x() == entry_zone_x && !force)
         return;
 
     m_promptItem->setPos(0, 0);
-    double x = 0 + m_promptItem->width() + HorizontalSpacing;
+    double x = std::max(0 + m_promptItem->width() + HorizontalSpacing, entry_zone_x);
     double y = 0;
     double width = 0;
 
@@ -1374,11 +1477,14 @@ void CommandEntry::startRemoving()
 
 WorksheetTextItem* CommandEntry::highlightItem()
 {
-    return m_commandItem;
+    return m_isExecutionEnabled ? m_commandItem : nullptr;
 }
 
 void CommandEntry::collapseResults()
 {
+    if (m_resultsCollapsed)
+        return;
+
     for(auto* item : m_resultItems) {
         fadeOutItem(item->graphicsObject(), nullptr);
         item->graphicsObject()->hide();
@@ -1400,6 +1506,9 @@ void CommandEntry::collapseResults()
 
 void CommandEntry::expandResults()
 {
+    if(!m_resultsCollapsed)
+        return;
+
     for(auto* item : m_resultItems) {
         fadeInItem(item->graphicsObject(), nullptr);
         item->graphicsObject()->show();
@@ -1439,3 +1548,44 @@ void CommandEntry::changeResultCollapsingAction()
     else
         collapseResults();
 }
+
+qreal CommandEntry::promptItemWidth()
+{
+    return m_promptItem->width();
+}
+
+void CommandEntry::excludeFromExecution()
+{
+    m_isExecutionEnabled = false;
+
+    KColorScheme scheme = KColorScheme(QPalette::Inactive, KColorScheme::View);
+
+    m_activeExecutionBackgroundColor = m_commandItem->backgroundColor();
+    m_activeExecutionTextColor = m_commandItem->defaultTextColor();
+
+    disconnect(m_commandItem, &WorksheetTextItem::receivedFocus, worksheet(), &Worksheet::highlightItem);
+
+    m_commandItem->setBackgroundColor(scheme.background(KColorScheme::AlternateBackground).color());
+    m_commandItem->setDefaultTextColor(scheme.foreground(KColorScheme::InactiveText).color());
+}
+
+void CommandEntry::addToExecution()
+{
+    m_isExecutionEnabled = true;
+
+    m_commandItem->setBackgroundColor(m_activeExecutionBackgroundColor);
+    m_commandItem->setDefaultTextColor(m_activeExecutionTextColor);
+
+    connect(m_commandItem, &WorksheetTextItem::receivedFocus, worksheet(), &Worksheet::highlightItem);
+    worksheet()->highlightItem(m_commandItem);
+}
+
+bool CommandEntry::isExcludedFromExecution()
+{
+    return m_isExecutionEnabled == false;
+}
+
+bool CommandEntry::isResultCollapsed()
+{
+    return m_resultsCollapsed;
+}
